(* Give me a filename and I'll give you a list of lists of values (as strings). *)
let read_csv filename =
	let rec input_lines file =
		match try [input_line file] with End_of_file -> [] with
		[] -> []
		| line -> (Str.split (Str.regexp_string ",") (List.nth line 0)) :: input_lines file in
	let file = open_in filename in
	input_lines file;;

(* converts data (a list of lists of strings) to a list of lists of floats *)
let floaterize data =
	let floaterize_elem elem =
		float_of_string elem in
	let floaterize_row row =
		List.map floaterize_elem row in
	List.map floaterize_row data;;

(* type definitions *)
(* leaf: nodenum value *)
(* split: nodenum feature, splitval, left, right *)
type decision_tree = Leaf of int * int | Split of int * int * float * decision_tree * decision_tree;;

let build_tree tree_data =
	(* nodeat returns a Leaf or Split for the item in tree_data referred to by nodenum *)
	let rec node_at tree_data nodenum =
		(* lookup looks in data (list of list of floats) for a list whose first value equals nodenum *)
		let rec lookup data nodenum =
			match data with
					[] -> failwith "Malformed tree. A node points to an invalid node id."
				| row :: rest -> if (List.hd row) = nodenum then row else lookup rest nodenum in
		let row = lookup tree_data nodenum in
		if (List.nth row 1) = 1.0   (* if it's a split node *)
			then begin
				let left = node_at tree_data (List.nth row 4) in
				let right = node_at tree_data (List.nth row 5) in
				let feature = int_of_float (List.nth row 2) in
				let nodenum_i = int_of_float nodenum in
				let splitval = List.nth row 3 in
				Split (nodenum_i, feature, splitval, left, right) end
			else Leaf (int_of_float nodenum, int_of_float (List.nth row 6)) in
	node_at tree_data (List.hd (List.hd tree_data));;

let rec query_tree tree wine_data =
	match tree with
		Leaf (_, value) -> value
	| Split (_, feature, splitval, left, right) ->
			begin
				let wine_value = List.nth wine_data feature in
				if wine_value <= splitval
					then query_tree left wine_data
					else query_tree right wine_data end;;

let print_decision decision row =
	print_int decision; print_string ",";
	let rec print_row row =
		match row with
			item :: [] -> begin
				print_string item;  print_endline ""; end
		|	item :: rest -> begin
				print_string item;  print_string ",";
				print_row rest end
		| [] -> print_string "" in  (* this last case should never be hit *)
	print_row (List.tl row);;  (* List.tl so that we don't print the actual score, only our decision *)

(* decisions is a list of ints, real_data is a list of lists of floats *)
let print_stats decisions real_data =
	let i_was_right decision row =
		if (float_of_int decision) = (List.nth row 0) then true else false in
	let right_or_wrong = List.map2 i_was_right decisions real_data in
	let kill_false item = item in
	let just_right = List.filter kill_false right_or_wrong in
	let num_right = List.length just_right in
	let num_total = List.length decisions in
	let num_wrong = num_total - num_right in
	Printf.eprintf "I missed %d out of %d for an accuracy of %.1f%%.\n"
		num_wrong
		num_total
		(((float_of_int num_right) /. (float_of_int num_total)) *. 100.);;

(* Check that there are precisely two arguments. *)
let num_args = Array.length Sys.argv in
if num_args != 3 then failwith "Requires two arguments, a decision tree and some data, both in CSV format.";;

(* Check that both arguments point to real files. *)
let ensure_exists filename =
	if not (Sys.file_exists filename) then failwith ("The file " ^ filename ^ " does not exist.") in
ensure_exists Sys.argv.(1);
ensure_exists Sys.argv.(2);;

let tree_file = Sys.argv.(1);;
let data_file = Sys.argv.(2);;



(* Construct the tree from tree_file, read in the data to query, and query. *)
let tree_data_as_strings = read_csv tree_file in
let tree_data = floaterize tree_data_as_strings in
let tree = build_tree tree_data in
let curried_query = query_tree tree in
let query_data_as_strings = read_csv data_file in
let query_data = floaterize query_data_as_strings in
let decisions = List.map curried_query query_data in
List.iter2 print_decision decisions query_data_as_strings;
print_stats decisions query_data;;